library ScrollByKeyboard;

uses
  Windows, Messages, SysUtils, RichEdit,
  AkelDLL_h, AkelEdit_h;

{$R ScrollByKeyboardDlg.res}
{$R *.RES}

const
   EM_EXLINEFROMCHAR = WM_USER + 54;
   EM_SCROLLCARET    = $00B7;        // from Messages.pas
//   EM_POSFROMCHAR    = WM_USER + 38;
//   EM_CHARFROMPOS    = WM_USER + 39;

  { TShortCut special values }
  scNone  = 0;
  scShift = $2000;
  scCtrl  = $4000;
  scAlt   = $8000;

  // Setting dialog consts:
  ID_LTEXT          = 101;
  ID_GROUPBOX       = 102;
  ID_CHECKBOX_CTRL  = 103;
  ID_CHECKBOX_ALT   = 104;
  ID_CHECKBOX_SHIFT = 105;
  ID_BTN_OK         = 1;
  ID_BTN_CANSEL     = 2;

type
  TShiftState = set of (ssShift, ssAlt, ssCtrl, ssLeft, ssRight, ssMiddle, ssDouble);

  TShortCut = Low(Word)..High(Word);

  TSmartScroll = class(TObject)
  private
    fMainWnd: HWND;
    fAkelEdit: Boolean;
    fOldWindows: Boolean;
    fPluginName: string;
    fwPluginName: WideString;
    fShift: TShiftState;
    fMainInitialized: Boolean;
    fhInstanceDLL: HINST;
    fMainIcon: HICON;
    fLangSystem: LANGID;
    procedure ReadOptionsA;
    procedure ReadOptionsW;
    procedure SaveOptionsA;
    procedure SaveOptionsW;
    function GetLangStringA(const wLangID: LANGID; const nStringID: Integer): PChar;
    function GetLangStringW(const wLangID: LANGID; const nStringID: Integer): PWideChar;
    procedure SetMainInitialized(const Value: Boolean);
  public
    NewEditProcData: PWndProcData;
    constructor Create(pd: TPluginData);
    destructor Destroy; override;
    property MainWnd: HWND read fMainWnd write fMainWnd;
    property hInstanceDLL: HINST read fhInstanceDLL write fhInstanceDLL;
    property MainIcon: HICON read fMainIcon write fMainIcon;
    property LangSystem: LANGID read fLangSystem write fLangSystem;
    property OldWindows: Boolean read fOldWindows write fOldWindows;
    property bAkelEdit: Boolean read fAkelEdit write fAkelEdit;
    property Shift: TShiftState read fShift write fShift;
    property MainInitialized: Boolean read fMainInitialized write SetMainInitialized;
  end;

var
  PSSmartScroll: TSmartScroll;
  PosX: Integer;
  IsDown: Boolean;

//  /:
procedure DLLEntryPoint(dwReason: DWord);
begin
  case dwReason of
    DLL_PROCESS_ATTACH:
      begin
      PosX:= -1;
      IsDown:= False;
      end;

    DLL_PROCESS_DETACH:
      if Assigned(PSSmartScroll) then
        FreeAndNil(PSSmartScroll);
        
    DLL_THREAD_ATTACH:  ;
    DLL_THREAD_DETACH:  ;
  end;
end;

function KeyToShortCut(Key: Word; Shift: TShiftState): TShortCut;
begin
  Result := 0;
  if WordRec(Key).Hi <> 0 then Exit;
  Result := Key;
  if ssShift in Shift then Inc(Result, scShift);
  if ssCtrl in Shift then Inc(Result, scCtrl);
  if ssAlt in Shift then Inc(Result, scAlt);
end;

function ShortCutToKey(ShortCut: TShortCut; var Key: Word): TShiftState;
begin
  Key := ShortCut and not (scShift + scCtrl + scAlt);
  Result := [];
  if ShortCut and scShift <> 0 then Include(Result, ssShift);
  if ShortCut and scCtrl <> 0 then Include(Result, ssCtrl);
  if ShortCut and scAlt <> 0 then Include(Result, ssAlt);
end;

function IsRequestedCtrls(const AKeyboardState: TKeyboardState;
  const AShift: TShiftState): Boolean;
//        
//  Ctrl, Alt  Shift.
// AKeyboardState -     
// AShift         -    
begin
  if (ssCtrl in AShift) then
    Result:= (AKeyboardState[VK_CONTROL] and $80 <> 0)
  else
    Result:= (AKeyboardState[VK_CONTROL] and $80 = 0);
  if not Result then
    Exit;

  if (ssAlt in AShift) then
    Result:= (AKeyboardState[VK_MENU] and $80 <> 0)
  else
    Result:= (AKeyboardState[VK_MENU] and $80 = 0);
  if not Result then
    Exit;

  if (ssShift in AShift) then
    Result:= (AKeyboardState[VK_SHIFT] and $80 <> 0)
  else
    Result:= (AKeyboardState[VK_SHIFT] and $80 = 0);
end;

procedure DllAkelPadID(var pv: TPluginVersion); cdecl; export;
begin
  pv.dwAkelDllVersion:= AKELDLL;
  pv.dwExeMinVersion3x:= MakeIdentifier(3,7,3,0);
  pv.dwExeMinVersion4x:= MakeIdentifier(4,3,1,0);
end;

function SettingDlgProcA(hDlg: HWND; Msg: UINT; wParam: WPARAM;
  lParam: LPARAM): LongBool; stdcall;
var
  ShiftState: TShiftState;
begin
  Result:= False;
  with PSSmartScroll do
  case Msg of
    WM_INITDIALOG:
    begin
      SendMessage(hDlg, WM_SETICON, ICON_BIG, MainIcon);
      SetDlgItemTextA(hDlg, ID_LTEXT, GetLangStringA(LangSystem, ID_LTEXT));
      SetDlgItemTextA(hDlg, ID_GROUPBOX, GetLangStringA(LangSystem, ID_GROUPBOX));
      SetDlgItemTextA(hDlg, ID_CHECKBOX_CTRL, GetLangStringA(LangSystem, ID_CHECKBOX_CTRL));
      SetDlgItemTextA(hDlg, ID_CHECKBOX_ALT, GetLangStringA(LangSystem, ID_CHECKBOX_ALT));
      SetDlgItemTextA(hDlg, ID_CHECKBOX_SHIFT, GetLangStringA(LangSystem, ID_CHECKBOX_SHIFT));
      SetDlgItemTextA(hDlg, ID_BTN_OK, GetLangStringA(LangSystem, ID_BTN_OK));
      SetDlgItemTextA(hDlg, ID_BTN_CANSEL, GetLangStringA(LangSystem, ID_BTN_CANSEL));

      if (ssCtrl in Shift) then
        SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_CTRL), BM_SETCHECK, BST_CHECKED, 0);
      if (ssAlt in Shift) then
        SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_ALT), BM_SETCHECK, BST_CHECKED, 0);
      if (ssShift in Shift) then
        SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_SHIFT), BM_SETCHECK, BST_CHECKED, 0);
    end;

    WM_COMMAND:
    begin
      if (LOWORD(wParam) = ID_BTN_OK) then
      begin
        ShiftState:= [];
        if (SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_CTRL), BM_GETCHECK, 0, 0) = BST_CHECKED) then
          ShiftState:= ShiftState+ [ssCtrl];
        if (SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_ALT), BM_GETCHECK, 0, 0) = BST_CHECKED) then
          ShiftState:= ShiftState+ [ssAlt];
        if (SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_SHIFT), BM_GETCHECK, 0, 0) = BST_CHECKED) then
          ShiftState:= ShiftState+ [ssShift];
        Shift:= ShiftState;
        SaveOptionsA;
        EndDialog(hDlg, 0);
        Result:= True;
      end
      else
      if (LOWORD(wParam) = ID_BTN_CANSEL) then
      begin
        EndDialog(hDlg, 0);
        Result:= True;
      end;
    end;

    WM_CLOSE:
    begin
      Windows.PostMessage(hDlg, WM_COMMAND, ID_BTN_CANSEL, 0);
      Result:= True;
    end;
  end;
end;

function SettingDlgProcW(hDlg: HWND; Msg: UINT; wParam: WPARAM;
  lParam: LPARAM): LongBool; stdcall;
var
  ShiftState: TShiftState;
begin
  Result:= False;
  with PSSmartScroll do
  case Msg of
    WM_INITDIALOG:
    begin
      SendMessage(hDlg, WM_SETICON, ICON_BIG, MainIcon);
      SetDlgItemTextW(hDlg, ID_LTEXT, GetLangStringW(LangSystem, ID_LTEXT));
      SetDlgItemTextW(hDlg, ID_GROUPBOX, GetLangStringW(LangSystem, ID_GROUPBOX));
      SetDlgItemTextW(hDlg, ID_CHECKBOX_CTRL, GetLangStringW(LangSystem, ID_CHECKBOX_CTRL));
      SetDlgItemTextW(hDlg, ID_CHECKBOX_ALT, GetLangStringW(LangSystem, ID_CHECKBOX_ALT));
      SetDlgItemTextW(hDlg, ID_CHECKBOX_SHIFT, GetLangStringW(LangSystem, ID_CHECKBOX_SHIFT));
      SetDlgItemTextW(hDlg, ID_BTN_OK, GetLangStringW(LangSystem, ID_BTN_OK));
      SetDlgItemTextW(hDlg, ID_BTN_CANSEL, GetLangStringW(LangSystem, ID_BTN_CANSEL));

      if (ssCtrl in Shift) then
        SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_CTRL), BM_SETCHECK, BST_CHECKED, 0);
      if (ssAlt in Shift) then
        SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_ALT), BM_SETCHECK, BST_CHECKED, 0);
      if (ssShift in Shift) then
        SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_SHIFT), BM_SETCHECK, BST_CHECKED, 0);
    end;

    WM_COMMAND:
    begin
      if (LOWORD(wParam) = ID_BTN_OK) then
      begin
        ShiftState:= [];
        if (SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_CTRL), BM_GETCHECK, 0, 0) = BST_CHECKED) then
          ShiftState:= ShiftState+ [ssCtrl];
        if (SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_ALT), BM_GETCHECK, 0, 0) = BST_CHECKED) then
          ShiftState:= ShiftState+ [ssAlt];
        if (SendMessage(GetDlgItem(hDlg, ID_CHECKBOX_SHIFT), BM_GETCHECK, 0, 0) = BST_CHECKED) then
          ShiftState:= ShiftState+ [ssShift];
        Shift:= ShiftState;
        SaveOptionsW;
        EndDialog(hDlg, 0);
        Result:= True;
      end
      else
      if (LOWORD(wParam) = ID_BTN_CANSEL) then
      begin
        EndDialog(hDlg, 0);
        Result:= True;
      end;
    end;

    WM_CLOSE:
    begin
      Windows.PostMessage(hDlg, WM_COMMAND, ID_BTN_CANSEL, 0);
      Result:= True;
    end;
  end;
end;

procedure Setting(var pd: TPluginData); cdecl; export;
const
  TemplateRes = 'SS_SETTING';
begin
  if (pd.lpbAutoLoad <> nil) then
  //   -  :
  begin
    pd.lpbAutoLoad^:= False;
    Exit;
  end;

  if not Assigned(PSSmartScroll) then
    PSSmartScroll:= TSmartScroll.Create(pd);

  with PSSmartScroll do
  if OldWindows then
	  DialogBoxA(hInstanceDLL, TemplateRes, MainWnd, @SettingDlgProcA)
  else
	  DialogBoxW(hInstanceDLL, TemplateRes, MainWnd, @SettingDlgProcW);

  //If plugin already loaded, stay in memory, but show as non-active
  if pd.bActive then
    pd.nUnload:= UD_NONUNLOAD_NONACTIVE;
end;

procedure Main(var pd: TPluginData); cdecl; export;
begin
  if (pd.lpbAutoLoad <> nil) then
  //   :
  begin
    pd.lpbAutoLoad^:= True;
    Exit;
  end;

  if not Assigned(PSSmartScroll) then
    PSSmartScroll:= TSmartScroll.Create(pd);
  with PSSmartScroll do
  begin
    MainInitialized:= not MainInitialized;
    if MainInitialized then
      //       :
      pd.nUnload:= UD_NONUNLOAD_ACTIVE;
  end;
end;

function NewEditProc(hWnd: HWND; uMsg, WParam, LParam: Longint): LRESULT; stdcall;
const
  Direction: array[Boolean] of Integer = (SB_LINEUP, SB_LINEDOWN); // array[VK_DOWN] is true
var
  KeyState: TKeyboardState;
  LineCount,
  FirstVisibleLine,         //       
  CurrentLine,              // ,  
  ScrollRes                 //      -       
    : Integer;

  function GetCaretPosEdt(const Handle: THandle; const IsMax: Boolean = True): TPoint;
  var
    CharRange: TCharRange;
  begin
    SendMessage(Handle, EM_EXGETSEL, 0, LongInt(@CharRange));
    if IsMax then
      Result.X := CharRange.cpMax
    else
      Result.X := CharRange.cpMin;
    Result.Y := SendMessage(Handle, EM_EXLINEFROMCHAR, 0, Result.X);
    Result.X := Result.X - SendMessage(Handle, EM_LINEINDEX, -1, 0);
  end;

  function GetLastVisibleLine(const IsAkelEdit: Boolean): Integer;
  var
     r: TRect;
     i: Integer;
  begin
    // The EM_GETRECT message retrieves the formatting rectangle of an edit control.
    if IsAkelEdit then
      SendMessage(hWnd, AEM_GETRECT, 0, Longint(@r))
    else
      SendMessage(hWnd, EM_GETRECT, 0, Longint(@r));
    r.Left:= r.Left + 1;
    r.Top:= r.Bottom - 2;

    // The EM_CHARFROMPOS message retrieves information about the character
    // closest to a specified point in the client area of an edit control
    i:= SendMessage(hWnd, EM_CHARFROMPOS, 0, Integer(@r.topleft));

    // The EM_EXLINEFROMCHAR message determines which
    // line contains the specified character in a rich edit control
    Result:= SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
  end;

  procedure AkelEditCommonProc(const CompareLine: Integer);
  // AkelEditCommonProc       
  //  /     
  //  AkelEdit.
  // CompareLine - ,     ,  
  //         (  
  //     ).
  var
    aes: TAESelection;
    ciCaret: TAECharIndex;
  begin
    if (CurrentLine = CompareLine) then
    begin
      //        ,   
      // ,      :
      if (SendMessage(hWnd, AEM_GETSEL, Integer(@ciCaret), Integer(@aes)) <> 0) then
      begin
        SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, Integer(True));
        try
          aes.dwFlags:= 0;
          aes.crSel.ciMin:= ciCaret;
          aes.crSel.ciMax:= ciCaret;
          SendMessage(hWnd, AEM_SETSEL, Integer(@ciCaret), Integer(@aes));
        finally
          SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, Integer(False));
        end;
      end;
    end;
    //  :
    SendMessage(hWnd, AEM_SCROLL, AESB_VERT, Direction[wParam = VK_DOWN]);
    if (CurrentLine = CompareLine) then
      //        /,  
      //    ,    /:
      SendMessage(hWnd, AEM_KEYDOWN, wParam, 0);
  end;

  procedure RichEditCommonProc(const CompareLine: Integer);
  // RichEdtCommonProc       
  //  /     
  //  RichEdit.
  // CompareLine - ,     ,  
  //         (  
  //     ).
  var
    LineLen, CurrPosY: Integer;
    CharRange: TCharRange;
    CaretPos: TPoint;
    Selection: TCharRange;
  begin
    if ((CurrentLine <= CompareLine) and (WParam = VK_DOWN)) or
       ((CurrentLine >= CompareLine) and (WParam = VK_UP)) then
    begin
      //     /   ,   
      // ,      :
      SendMessage(hWnd, EM_EXGETSEL, 0, Longint(@Selection));
      if (Selection.cpMin <> Selection.cpMax) then
      begin
        Selection.cpMin:= Selection.cpMax;
        SendMessage(hWnd, EM_EXSETSEL, 0, Longint(@Selection));
      end;
    end;

    //  :
    SendMessage(hWnd, EM_SCROLL, Direction[wParam = VK_DOWN], 0);

    if ((CurrentLine <= CompareLine) and (WParam = VK_DOWN)) or
       ((CurrentLine >= CompareLine) and (WParam = VK_UP)) then
    //        /,  
    //    ,     ( 
    //    ,  ):
    begin
      if (PosX = -1) then
        //         ( 
        //   PosX    -1),    
        //         , 
        //  ,      :
        PosX:= GetCaretPosEdt(hWnd).X;

      //    (      ):
      CaretPos:= GetCaretPosEdt(hWnd);
      //    :
      if (WParam = VK_DOWN) then
        Inc(CaretPos.Y)
      else
        Dec(CaretPos.Y);
      //      :
      CurrPosY:= SendMessage(hWnd, EM_LINEINDEX, CaretPos.Y, 0);
      //   :
      LineLen:= SendMessage(hWnd, EM_LINELENGTH, CurrPosY, 0);
      //     :
      if PosX < LineLen then
        CaretPos.X:= PosX
      else
        CaretPos.X:= LineLen;
      //
      CharRange.cpMin:= SendMessage(hWnd, EM_LINEINDEX, CaretPos.y, 0)+ CaretPos.x;
      CharRange.cpMax:= CharRange.cpMin;
      SendMessage(hWnd, EM_EXSETSEL, 0, Longint(@CharRange));
    end;
  end;

begin
  with PSSmartScroll do
  case uMsg of
    WM_KEYUP, WM_SYSKEYUP:
      begin
        if not (wParam in [VK_DOWN, VK_UP]) then
        begin
          PosX:= -1;
          IsDown:= False;
        end;
        Result:= NewEditProcData.NextProc(hWnd, uMsg, WParam, LParam);
        Exit;
      end;

    WM_KEYDOWN, WM_SYSKEYDOWN:
      begin
        GetKeyboardState(KeyState);
        if not ((wParam in [VK_DOWN, VK_UP]) and IsRequestedCtrls(KeyState, Shift)) then
        //       -    
        //     :
        begin
          Result:= NewEditProcData.NextProc(hWnd, uMsg, WParam, LParam);
          Exit;
        end
        else
          Result:= 0;

        FirstVisibleLine:= SendMessage(hWnd, EM_GETFIRSTVISIBLELINE, 0, 0);
        LineCount:= SendMessage(hWnd, EM_GETLINECOUNT, 0, 0);
        //       ,    :
        if (FirstVisibleLine = 0) and (GetLastVisibleLine(bAkelEdit) = LineCount- 1) then
          Exit;

        if bAkelEdit then
        {$REGION 'AkelPad v.4.x.x'}
        begin
          //        ,
          //    ,       
          //    ,     :
          ScrollRes:= SendMessage(hWnd, AEM_SCROLLCARET,
                              AESC_UNITCHARX or AESC_UNITCHARY, MakeLong(0, 0));
          if (ScrollRes and AECSE_SCROLLEDY = AECSE_SCROLLEDY) then
            Exit;

          CurrentLine:= SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, -1);
          case wParam of
            VK_DOWN: //  :
              AkelEditCommonProc(SendMessage(hWnd, EM_GETFIRSTVISIBLELINE, 0, 0));

            VK_UP: //  :
              AkelEditCommonProc(GetLastVisibleLine(True));
          end;
        end
        {$ENDREGION}
        else
        // AkelPad v.3.x.x:
        begin
          //        ,
          //    ,       
          //    ,     :
          if not IsDown then
          begin
            IsDown:= True;
            SendMessage(hWnd, EM_SCROLLCARET, 0, 0);
            if (FirstVisibleLine <> SendMessage(hWnd, EM_GETFIRSTVISIBLELINE, 0, 0)) then
              Exit;
          end;

          CurrentLine:= GetCaretPosEdt(hWnd).Y;
          case wParam of
            VK_DOWN: //  :
              RichEditCommonProc(SendMessage(hWnd, EM_GETFIRSTVISIBLELINE, 0, 0));

            VK_UP: //  :
              RichEditCommonProc(GetLastVisibleLine(False)- 1);
          end;
        end;
      end;
  else
    Result:= NewEditProcData.NextProc(hWnd, uMsg, WParam, LParam);
  end;
end;

{ TSmartScroll }

constructor TSmartScroll.Create(pd: TPluginData);
var
  i: Integer;
begin
  with pd do
  begin
    fMainWnd:= hMainWnd;
    fOldWindows:= bOldWindows;
    fAkelEdit:= bAkelEdit;
    fhInstanceDLL:= hInstanceDLL;
    fMainIcon:= hMainIcon;
    fLangSystem:= PRIMARYLANGID(wLangSystem);
  end;

  if fOldWindows then
  begin
    fPluginName:= PChar(pd.pFunction);
    i:= 0;
    while PChar(pd.pFunction)[i] <> ':' do
      Inc(i);
    SetLength(fPluginName, i);
    ReadOptionsA;
  end
  else
  begin
    fwPluginName:= PWideChar(pd.pFunction);
    i:= 0;
    while PWideChar(pd.pFunction)[i] <> ':' do
      Inc(i);
    SetLength(fwPluginName, i);
    ReadOptionsW;
  end;
  fMainInitialized:= False;
  NewEditProcData:= nil;
end;

destructor TSmartScroll.Destroy;
begin
  MainInitialized:= False;
  inherited;
end;

function TSmartScroll.GetLangStringA(const wLangID: LANGID;
  const nStringID: Integer): PChar;
begin
  Result:= nil;
  if (wLangID = LANG_RUSSIAN) then
  begin
    case nStringID of
      ID_LTEXT         : Result:= '  -, '+
                                  '    Up  '+
                                  'Down  ';
      ID_GROUPBOX      : Result:= '-';
      ID_CHECKBOX_CTRL : Result:= 'CTRL';
      ID_CHECKBOX_ALT  : Result:= 'ALT';
      ID_CHECKBOX_SHIFT: Result:= 'SHIFT';
      ID_BTN_OK        : Result:= 'OK';
      ID_BTN_CANSEL    : Result:= '';
    end;
  end
  else
  begin
    case nStringID of
      ID_LTEXT         : Result:= 'Define a set of control keys for initialization '+
                                  'together with keys Up or Down a mode of '+
                                  'scrolling of the text';
      ID_GROUPBOX      : Result:= 'Control keys';
      ID_CHECKBOX_CTRL : Result:= 'CTRL';
      ID_CHECKBOX_ALT  : Result:= 'ALT';
      ID_CHECKBOX_SHIFT: Result:= 'SHIFT';
      ID_BTN_OK        : Result:= 'OK';
      ID_BTN_CANSEL    : Result:= 'Cancel';
    end;
  end;
end;

function TSmartScroll.GetLangStringW(const wLangID: LANGID;
  const nStringID: Integer): PWideChar;
begin
  Result:= nil;
  if (wLangID = LANG_RUSSIAN) then
  begin
    case nStringID of
      ID_LTEXT         : Result:=
        #1054#1087#1088#1077#1076#1077#1083#1080#1090#1077' '+
        #1085#1072#1073#1086#1088' '#1082#1083#1072#1074#1080#1096'-'+
        #1084#1086#1076#1080#1092#1080#1082#1072#1090#1086#1088#1086#1074', '+
        #1080#1085#1080#1094#1080#1080#1088#1091#1102#1097#1080#1093' '+
        #1089#1086#1074#1084#1077#1089#1090#1085#1086' '#1089' ' +
        #1082#1083#1072#1074#1080#1096#1072#1084#1080' Up '+
        #1080#1083#1080' Down '#1089#1082#1088#1086#1083#1083#1080#1085#1075' '+
        #1090#1077#1082#1089#1090#1072;
      ID_GROUPBOX      : Result:=
        #1050#1083#1072#1074#1080#1096#1080'-'+
        #1084#1086#1076#1080#1092#1080#1082#1072#1090#1086#1088#1099;
      ID_CHECKBOX_CTRL : Result:= 'CTRL';
      ID_CHECKBOX_ALT  : Result:= 'ALT';
      ID_CHECKBOX_SHIFT: Result:= 'SHIFT';
      ID_BTN_OK        : Result:= 'OK';
      ID_BTN_CANSEL    : Result:= #1054#1090#1084#1077#1085#1072; 
    end;
  end
  else
  begin
    case nStringID of
      ID_LTEXT         : Result:= 'Define a set of control keys for initialization '+
                                  'together with keys Up or Down a mode of '+
                                  'scrolling of the text';
      ID_GROUPBOX      : Result:= 'Control keys';
      ID_CHECKBOX_CTRL : Result:= 'CTRL';
      ID_CHECKBOX_ALT  : Result:= 'ALT';
      ID_CHECKBOX_SHIFT: Result:= 'SHIFT';
      ID_BTN_OK        : Result:= 'OK';
      ID_BTN_CANSEL    : Result:= 'Cancel';
    end;
  end;
end;

procedure TSmartScroll.ReadOptionsA;
var
  PluginOption: TPluginOptionA;
  hOptions: THandle;
  ShortCut: TShortCut;
  Key: Word;
begin
  hOptions:= SendMessage(MainWnd, AKD_BEGINOPTIONS, POB_READ, LPARAM(fPluginName));
  if (hOptions <> 0) then
  with PluginOption do
  try
    szOptionName:= 'ScrollCtrls';
    dwType:= PO_DWORD;
    lpData:= PByte(@ShortCut);
    dwData:= SizeOf(TShortCut);
    SendMessage(MainWnd, AKD_OPTION, hOptions, LPARAM(@PluginOption));
    Shift:= ShortCutToKey(ShortCut, key);
  finally
    SendMessage(MainWnd, AKD_ENDOPTIONS, hOptions, 0);
  end
  else
  begin
    Shift:= [ssCtrl];
    SaveOptionsW;
  end;
end;

procedure TSmartScroll.ReadOptionsW;
var
  PluginOption: TPluginOptionW;
  hOptions: THandle;
  ShortCut: TShortCut;
  Key: Word;
begin
  hOptions:= SendMessage(MainWnd, AKD_BEGINOPTIONS, POB_READ, LPARAM(fwPluginName));
  if (hOptions <> 0) then
  try
    PluginOption.wszOptionName:= 'ScrollCtrls';
    PluginOption.dwType:= PO_DWORD;
    PluginOption.lpData:= PByte(@ShortCut);
    PluginOption.dwData:= SizeOf(TShortCut);
    SendMessage(MainWnd, AKD_OPTION, hOptions, LPARAM(@PluginOption));
    Shift:= ShortCutToKey(ShortCut, key);
  finally
    SendMessage(MainWnd, AKD_ENDOPTIONS, hOptions, 0);
  end
  else
  begin
    Shift:= [ssCtrl];
    SaveOptionsW;
  end;
end;

procedure TSmartScroll.SaveOptionsA;
var
  PluginOption: TPluginOptionA;
  hOptions: THandle;
  ShortCut: TShortCut;
begin
  hOptions:= SendMessage(MainWnd, AKD_BEGINOPTIONS, POB_SAVE, LPARAM(fPluginName));
  if (hOptions <> 0) then
  try
    ShortCut:= KeyToShortCut(0, Shift);
    PluginOption.szOptionName:= 'ScrollCtrls';
    PluginOption.dwType:= PO_DWORD;
    PluginOption.lpData:= PByte(@ShortCut);
    PluginOption.dwData:= SizeOf(ShortCut);
    SendMessage(MainWnd, AKD_OPTION, hOptions, LPARAM(@PluginOption));
  finally
    SendMessage(MainWnd, AKD_ENDOPTIONS, hOptions, 0);
  end;
end;

procedure TSmartScroll.SaveOptionsW;
var
  PluginOption: TPluginOptionW;
  hOptions: THandle;
  ShortCut: TShortCut;
begin
  hOptions:= SendMessage(MainWnd, AKD_BEGINOPTIONS, POB_SAVE, LPARAM(fwPluginName));
  if (hOptions <> 0) then
  try
    ShortCut:= KeyToShortCut(0, Shift);
    PluginOption.wszOptionName:= 'ScrollCtrls';
    PluginOption.dwType:= PO_DWORD;
    PluginOption.lpData:= PByte(@ShortCut);
    PluginOption.dwData:= SizeOf(ShortCut);
    SendMessage(MainWnd, AKD_OPTION, hOptions, LPARAM(@PluginOption));
  finally
    SendMessage(MainWnd, AKD_ENDOPTIONS, hOptions, 0);
  end;
end;

procedure TSmartScroll.SetMainInitialized(const Value: Boolean);
begin
  if (fMainInitialized <> Value) then
  begin
    fMainInitialized:= Value;
    if Value then
      //SubClass edit procedure
      SendMessage(MainWnd, AKD_SETEDITPROC, WPARAM(@NewEditProc), LPARAM(@NewEditProcData))
    else
    begin
      SendMessage(MainWnd, AKD_SETEDITPROC, 0, LPARAM(@NewEditProcData));
      NewEditProcData:= nil;
    end;
  end;
end;

exports
  DllAkelPadID,
  Main,
  Setting;

begin
  DllProc:= @DLLEntryPoint;
end.
